@@ -34,6 +34,8 @@
 #include <string.h>
 #include <sys/types.h>
 #include <assert.h>
+#include <unistd.h>
+#include <locale.h>
 #include <fcntl.h> /* FIXME */
 
 #include "rungpg.h"
@@ -46,6 +48,7 @@
 #include "wait.h"
 #include "io.h"
 #include "key.h"
+#include "sema.h"
 
 #include "engine-gpgsm.h"
 
@@ -56,20 +59,30 @@
 #define xtoi_2(p)   ((xtoi_1(p) * 16) + xtoi_1((p)+1))
 
 
+
+typedef struct
+{
+  int fd;	/* FD we talk about.  */
+  int dir;	/* Inbound/Outbound, maybe given implicit?  */
+  void *data;	/* Handler-specific data.  */
+  void *tag;	/* ID from the user for gpgme_remove_io_callback.  */
+} iocb_data_t;
+
 struct gpgsm_object_s
 {
   ASSUAN_CONTEXT assuan_ctx;
 
+  iocb_data_t status_cb;
+
   /* Input, output etc are from the servers perspective.  */
-  int input_fd;
+  iocb_data_t input_cb;
   int input_fd_server;
-  GpgmeData input_data;
-  int output_fd;
+
+  iocb_data_t output_cb;
   int output_fd_server;
-  GpgmeData output_data;
-  int message_fd;
+
+  iocb_data_t message_cb;
   int message_fd_server;
-  GpgmeData message_data;
 
   char *command;
 
@@ -90,6 +103,8 @@ struct gpgsm_object_s
       int linelen;
     } attic;
   } colon; 
+
+  struct GpgmeIOCbs io_cbs;
 };
 
 
@@ -97,10 +112,12 @@ const char *
 _gpgme_gpgsm_get_version (void)
 {
   static const char *gpgsm_version;
+  DEFINE_STATIC_LOCK (gpgsm_version_lock);
 
-  /* FIXME: Locking.  */
+  LOCK (gpgsm_version_lock);
   if (!gpgsm_version)
     gpgsm_version = _gpgme_get_program_version (_gpgme_get_gpgsm_path ());
+  UNLOCK (gpgsm_version_lock);
 
   return gpgsm_version;
 }
@@ -119,133 +136,49 @@ static void
 close_notify_handler (int fd, void *opaque)
 {
   GpgsmObject gpgsm = opaque;
+  int possibly_done = 0;
 
   assert (fd != -1);
-  if (gpgsm->input_fd == fd)
-    gpgsm->input_fd = -1;
-  else if (gpgsm->output_fd == fd)
-    gpgsm->output_fd = -1;
-  else if (gpgsm->message_fd == fd)
-    gpgsm->message_fd = -1;
-}
-
-
-GpgmeError
-_gpgme_gpgsm_new (GpgsmObject *r_gpgsm)
-{
-  GpgmeError err = 0;
-  GpgsmObject gpgsm;
-  char *argv[] = { "gpgsm", "--server", NULL };
-  int fds[2];
-  int child_fds[4];
-
-  *r_gpgsm = NULL;
-  gpgsm = xtrycalloc (1, sizeof *gpgsm);
-  if (!gpgsm)
+  if (gpgsm->status_cb.fd == fd)
     {
-      err = mk_error (Out_Of_Core);
-      return err;
+      if (gpgsm->status_cb.tag)
+	{
+	  (*gpgsm->io_cbs.remove) (gpgsm->status_cb.tag);
+	  possibly_done = 1;
+	}
+      gpgsm->status_cb.fd = -1;
     }
-
-  gpgsm->input_fd = -1;
-  gpgsm->input_fd_server = -1;
-  gpgsm->output_fd = -1;
-  gpgsm->output_fd_server = -1;
-  gpgsm->message_fd = -1;
-  gpgsm->message_fd_server = -1;
-
-  gpgsm->status.fnc = 0;
-  gpgsm->colon.fnc = 0;
-  gpgsm->colon.attic.line = 0;
-  gpgsm->colon.attic.linesize = 0;
-  gpgsm->colon.attic.linelen = 0;
-
-  if (_gpgme_io_pipe (fds, 0) < 0)
+  else if (gpgsm->input_cb.fd == fd)
     {
-      err = mk_error (Pipe_Error);
-      goto leave;
+      if (gpgsm->input_cb.tag)
+	{
+	  (*gpgsm->io_cbs.remove) (gpgsm->input_cb.tag);
+	  possibly_done = 1;
+	}
+      gpgsm->input_cb.fd = -1;
     }
-  gpgsm->input_fd = fds[1];
-  gpgsm->input_fd_server = fds[0];
-
-  if (_gpgme_io_pipe (fds, 1) < 0)
+  else if (gpgsm->output_cb.fd == fd)
     {
-      err = mk_error (Pipe_Error);
-      goto leave;
+      if (gpgsm->output_cb.tag)
+	{      
+	  (*gpgsm->io_cbs.remove) (gpgsm->output_cb.tag);
+	  possibly_done = 1;
+	}
+      gpgsm->output_cb.fd = -1;
     }
-  gpgsm->output_fd = fds[0];
-  gpgsm->output_fd_server = fds[1];
-
-  if (_gpgme_io_pipe (fds, 0) < 0)
+  else if (gpgsm->message_cb.fd == fd)
     {
-      err = mk_error (Pipe_Error);
-      goto leave;
+      if (gpgsm->message_cb.tag)
+	{
+	  (*gpgsm->io_cbs.remove) (gpgsm->message_cb.tag);
+	  possibly_done = 1;
+	}
+      gpgsm->message_cb.fd = -1;
     }
-  gpgsm->message_fd = fds[1];
-  gpgsm->message_fd_server = fds[0];
-
-  child_fds[0] = gpgsm->input_fd_server;
-  child_fds[1] = gpgsm->output_fd_server;
-  child_fds[2] = gpgsm->message_fd_server;
-  child_fds[3] = -1;
-  err = assuan_pipe_connect (&gpgsm->assuan_ctx,
-			     _gpgme_get_gpgsm_path (), argv, child_fds);
-
-  if (!err &&
-      (_gpgme_io_set_close_notify (gpgsm->input_fd,
-				   close_notify_handler, gpgsm)
-       || _gpgme_io_set_close_notify (gpgsm->output_fd,
-				      close_notify_handler, gpgsm)
-       || _gpgme_io_set_close_notify (gpgsm->message_fd,
-				      close_notify_handler, gpgsm)))
-    {
-      err = mk_error (General_Error);
-      goto leave;
-    }
-      
- leave:
-  /* Close the server ends of the pipes.  Our ends are closed in
-     _gpgme_gpgsm_release.  */
-  if (gpgsm->input_fd_server != -1)
-    _gpgme_io_close (gpgsm->input_fd_server);
-  if (gpgsm->output_fd_server != -1)
-    _gpgme_io_close (gpgsm->output_fd_server);
-  if (gpgsm->message_fd_server != -1)
-    _gpgme_io_close (gpgsm->message_fd_server);
-
-  if (err)
-    _gpgme_gpgsm_release (gpgsm);
-  else
-    *r_gpgsm = gpgsm;
-
-  return err;
-}
-
-
-void
-_gpgme_gpgsm_release (GpgsmObject gpgsm)
-{
-  pid_t pid;
-
-  if (!gpgsm)
-    return;
-
-  pid = assuan_get_pid (gpgsm->assuan_ctx);
-  if (pid != -1)
-    _gpgme_remove_proc_from_wait_queue (pid);
-
-  if (gpgsm->input_fd != -1)
-    _gpgme_io_close (gpgsm->input_fd);
-  if (gpgsm->output_fd != -1)
-    _gpgme_io_close (gpgsm->output_fd);
-  if (gpgsm->message_fd != -1)
-    _gpgme_io_close (gpgsm->message_fd);
-
-  assuan_disconnect (gpgsm->assuan_ctx);
-
-  xfree (gpgsm->colon.attic.line);
-  xfree (gpgsm->command);
-  xfree (gpgsm);
+  if (possibly_done && gpgsm->io_cbs.event
+      && gpgsm->status_cb.fd == -1 && gpgsm->input_cb.fd == -1
+      && gpgsm->output_cb.fd == -1 && gpgsm->message_cb.fd == -1)
+    (*gpgsm->io_cbs.event) (gpgsm->io_cbs.event_priv, GPGME_EVENT_DONE, NULL);
 }
 
 
@@ -279,14 +212,14 @@ map_assuan_error (AssuanError err)
     case ASSUAN_Accept_Failed:
       return mk_error (General_Error);
 
-  /* The following error codes are meant as status codes.  */
+      /* The following error codes are meant as status codes.  */
     case ASSUAN_Not_Implemented:
       return mk_error (Not_Implemented);
     case ASSUAN_Canceled:
       return mk_error (Canceled);
     case ASSUAN_Unsupported_Algorithm:
       return mk_error (Not_Implemented);  /* XXX Argh.  */
-
+      
       /* These are errors internal to GPGME.  */
     case ASSUAN_No_Data_Available:
     case ASSUAN_No_Input:
@@ -304,6 +237,11 @@ map_assuan_error (AssuanError err)
     case ASSUAN_Inquire_Unknown:
     case ASSUAN_Inquire_Error:
     case ASSUAN_Invalid_Option:
+    case ASSUAN_Invalid_Index:
+    case ASSUAN_Unexpected_Status:
+    case ASSUAN_Unexpected_Data:
+    case ASSUAN_Invalid_Status:
+    case ASSUAN_Not_Confirmed:
       return mk_error (General_Error);
 
       /* These are errors in the server.  */
@@ -321,16 +259,21 @@ map_assuan_error (AssuanError err)
     case ASSUAN_No_Public_Key:
     case ASSUAN_No_Secret_Key:
     case ASSUAN_Invalid_Name:
-      return mk_error(Invalid_Key);
+    case ASSUAN_Card_Error:	/* XXX: Oh well.  */
+    case ASSUAN_Invalid_Card:	/* XXX: Oh well.  */
+    case ASSUAN_No_PKCS15_App:	/* XXX: Oh well.  */
+    case ASSUAN_Card_Not_Present:	/* XXX: Oh well.  */
+    case ASSUAN_Invalid_Id:	/* XXX: Oh well.  */
+      return mk_error (Invalid_Key);
 
     case ASSUAN_Bad_Signature:
-      return mk_error(Invalid_Key);  /* XXX: This is wrong.  */
+      return mk_error (Invalid_Key);  /* XXX: This is wrong.  */
 
     case ASSUAN_Cert_Revoked:
     case ASSUAN_No_CRL_For_Cert:
     case ASSUAN_CRL_Too_Old:
     case ASSUAN_Not_Trusted:
-      return mk_error(Invalid_Key);  /* XXX Some more details would be good.  */
+      return mk_error (Invalid_Key);  /* XXX Some more details would be good.  */
 
     default:
       return mk_error (General_Error);
@@ -338,8 +281,273 @@ map_assuan_error (AssuanError err)
 }
 
 
+GpgmeError
+_gpgme_gpgsm_new (GpgsmObject *r_gpgsm)
+{
+  GpgmeError err = 0;
+  GpgsmObject gpgsm;
+  char *argv[3];
+  int fds[2];
+  int child_fds[4];
+  char *dft_display = NULL;
+  char *dft_ttyname = NULL;
+  char *dft_ttytype = NULL;
+  char *old_lc = NULL;
+  char *dft_lc = NULL;
+  char *optstr;
+  int fdlist[5];
+  int nfds;
+
+  *r_gpgsm = NULL;
+  gpgsm = xtrycalloc (1, sizeof *gpgsm);
+  if (!gpgsm)
+    {
+      err = mk_error (Out_Of_Core);
+      return err;
+    }
+
+  gpgsm->status_cb.fd = -1;
+  gpgsm->status_cb.tag = 0;
+
+  gpgsm->input_cb.fd = -1;
+  gpgsm->input_cb.tag = 0;
+  gpgsm->input_fd_server = -1;
+  gpgsm->output_cb.fd = -1;
+  gpgsm->output_cb.tag = 0;
+  gpgsm->output_fd_server = -1;
+  gpgsm->message_cb.fd = -1;
+  gpgsm->message_cb.tag = 0;
+  gpgsm->message_fd_server = -1;
+
+  gpgsm->status.fnc = 0;
+  gpgsm->colon.fnc = 0;
+  gpgsm->colon.attic.line = 0;
+  gpgsm->colon.attic.linesize = 0;
+  gpgsm->colon.attic.linelen = 0;
+
+  gpgsm->io_cbs.add = NULL;
+  gpgsm->io_cbs.add_priv = NULL;
+  gpgsm->io_cbs.remove = NULL;
+  gpgsm->io_cbs.event = NULL;
+  gpgsm->io_cbs.event_priv = NULL;
+
+  if (_gpgme_io_pipe (fds, 0) < 0)
+    {
+      err = mk_error (Pipe_Error);
+      goto leave;
+    }
+  gpgsm->input_cb.fd = fds[1];
+  gpgsm->input_cb.dir = 0;
+  gpgsm->input_fd_server = fds[0];
+
+  if (_gpgme_io_pipe (fds, 1) < 0)
+    {
+      err = mk_error (Pipe_Error);
+      goto leave;
+    }
+  gpgsm->output_cb.fd = fds[0];
+  gpgsm->output_cb.dir = 1;
+  gpgsm->output_fd_server = fds[1];
+
+  if (_gpgme_io_pipe (fds, 0) < 0)
+    {
+      err = mk_error (Pipe_Error);
+      goto leave;
+    }
+  gpgsm->message_cb.fd = fds[1];
+  gpgsm->message_cb.dir = 0;
+  gpgsm->message_fd_server = fds[0];
+
+  child_fds[0] = gpgsm->input_fd_server;
+  child_fds[1] = gpgsm->output_fd_server;
+  child_fds[2] = gpgsm->message_fd_server;
+  child_fds[3] = -1;
+
+  argv[0] = "gpgsm";
+  argv[1] = "--server";
+  argv[2] = NULL;
+
+  err = assuan_pipe_connect2 (&gpgsm->assuan_ctx,
+                              _gpgme_get_gpgsm_path (), argv, child_fds,
+                              1 /* dup stderr to /dev/null */);
+
+  /* We need to know the fd used by assuan for reads.  We do this by
+     using the assumption that the first returned fd from
+     assuan_get_active_fds() is always this one.  */
+  nfds = assuan_get_active_fds (gpgsm->assuan_ctx, 0 /* read fds */,
+                                fdlist, DIM (fdlist));
+  if (nfds < 1)
+    {
+      err = mk_error (General_Error);  /* FIXME */
+      goto leave;
+    }
+  /* We duplicate the file descriptor, so we can close it without
+     disturbing assuan.  Alternatively, we could special case
+     status_fd and register/unregister it manually as needed, but this
+     increases code duplication and is more complicated as we can not
+     use the close notifications etc.  */
+  gpgsm->status_cb.fd = dup (fdlist[0]);
+  if (gpgsm->status_cb.fd < 0)
+    {
+      err = mk_error (General_Error);	/* FIXME */
+      goto leave;
+    }
+  gpgsm->status_cb.dir = 1;
+  gpgsm->status_cb.data = gpgsm;
+
+  dft_display = getenv ("DISPLAY");
+  if (dft_display)
+    {
+      if (asprintf (&optstr, "OPTION display=%s", dft_display) < 0)
+        {
+	  err = mk_error (Out_Of_Core);
+	  goto leave;
+	}
+      err = assuan_transact (gpgsm->assuan_ctx, optstr, NULL, NULL, NULL,
+			     NULL, NULL, NULL);
+      free (optstr);
+      if (err)
+	{
+	  err = map_assuan_error (err);
+	  goto leave;
+	}
+    }
+  dft_ttyname = ttyname (1);
+  if (dft_ttyname)
+    {
+      if (asprintf (&optstr, "OPTION ttyname=%s", dft_ttyname) < 0)
+        {
+	  err = mk_error (Out_Of_Core);
+	  goto leave;
+	}
+      err = assuan_transact (gpgsm->assuan_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
+			     NULL);
+      free (optstr);
+      if (err)
+	{
+	  err = map_assuan_error (err);
+	  goto leave;
+	}
+
+      dft_ttytype = getenv ("TERM");
+      if (dft_ttytype)
+	{
+	  if (asprintf (&optstr, "OPTION ttytype=%s", dft_ttytype) < 0)
+	    {
+	      err = mk_error (Out_Of_Core);
+	      goto leave;
+	    }
+	  err = assuan_transact (gpgsm->assuan_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
+				 NULL);
+	  free (optstr);
+	  if (err)
+	    {
+	      err = map_assuan_error (err);
+	      goto leave;
+	    }
+	}
+      old_lc = setlocale (LC_CTYPE, NULL);
+      dft_lc = setlocale (LC_CTYPE, "");
+      if (dft_lc)
+	{
+	  if (asprintf (&optstr, "OPTION lc-ctype=%s", dft_lc) < 0)
+	    err = mk_error (Out_Of_Core);
+	  else
+	    {
+	      err = assuan_transact (gpgsm->assuan_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
+				     NULL);
+	      free (optstr);
+	      if (err)
+		err = map_assuan_error (err);
+	    }
+	}
+      if (old_lc)
+	setlocale (LC_CTYPE, old_lc);
+      if (err)
+	goto leave;
+
+      old_lc = setlocale (LC_MESSAGES, NULL);
+      dft_lc = setlocale (LC_MESSAGES, "");
+      if (dft_lc)
+	{
+	  if (asprintf (&optstr, "OPTION lc-messages=%s", dft_lc) < 0)
+	    err = mk_error (Out_Of_Core);
+	  else
+	    {
+	      err = assuan_transact (gpgsm->assuan_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
+				     NULL);
+	      free (optstr);
+	      if (err)
+		err = map_assuan_error (err);
+	    }
+	}
+      if (old_lc)
+	setlocale (LC_MESSAGES, old_lc);
+      if (err)
+	goto leave;
+    }
+
+  if (!err &&
+      (_gpgme_io_set_close_notify (gpgsm->status_cb.fd,
+				   close_notify_handler, gpgsm)
+       || _gpgme_io_set_close_notify (gpgsm->input_cb.fd,
+				   close_notify_handler, gpgsm)
+       || _gpgme_io_set_close_notify (gpgsm->output_cb.fd,
+				      close_notify_handler, gpgsm)
+       || _gpgme_io_set_close_notify (gpgsm->message_cb.fd,
+				      close_notify_handler, gpgsm)))
+    {
+      err = mk_error (General_Error);
+      goto leave;
+    }
+      
+ leave:
+  /* Close the server ends of the pipes.  Our ends are closed in
+     _gpgme_gpgsm_release.  */
+  if (gpgsm->input_fd_server != -1)
+    _gpgme_io_close (gpgsm->input_fd_server);
+  if (gpgsm->output_fd_server != -1)
+    _gpgme_io_close (gpgsm->output_fd_server);
+  if (gpgsm->message_fd_server != -1)
+    _gpgme_io_close (gpgsm->message_fd_server);
+
+  if (err)
+    _gpgme_gpgsm_release (gpgsm);
+  else
+    *r_gpgsm = gpgsm;
+
+  return err;
+}
+
+
+void
+_gpgme_gpgsm_release (GpgsmObject gpgsm)
+{
+  if (!gpgsm)
+    return;
+
+  if (gpgsm->status_cb.fd != -1)
+    _gpgme_io_close (gpgsm->status_cb.fd);
+  if (gpgsm->input_cb.fd != -1)
+    _gpgme_io_close (gpgsm->input_cb.fd);
+  if (gpgsm->output_cb.fd != -1)
+    _gpgme_io_close (gpgsm->output_cb.fd);
+  if (gpgsm->message_cb.fd != -1)
+    _gpgme_io_close (gpgsm->message_cb.fd);
+
+  assuan_disconnect (gpgsm->assuan_ctx);
+
+  xfree (gpgsm->colon.attic.line);
+  xfree (gpgsm->command);
+  xfree (gpgsm);
+}
+
+/* Forward declaration.  */
+static GpgStatusCode parse_status (const char *name);
+
 static GpgmeError
-gpgsm_assuan_simple_command (ASSUAN_CONTEXT ctx, char *cmd)
+gpgsm_assuan_simple_command (ASSUAN_CONTEXT ctx, char *cmd, GpgStatusHandler status_fnc,
+			     void *status_fnc_value)
 {
   AssuanError err;
   char *line;
@@ -354,22 +562,43 @@ gpgsm_assuan_simple_command (ASSUAN_CONTEXT ctx, char *cmd)
       err = assuan_read_line (ctx, &line, &linelen);
       if (err)
 	return map_assuan_error (err);
+
+      if (*line == '#' || !linelen)
+	continue;
+
+      if (linelen >= 2
+	  && line[0] == 'O' && line[1] == 'K'
+	  && (line[2] == '\0' || line[2] == ' '))
+	return 0;
+      else if (linelen >= 4
+	  && line[0] == 'E' && line[1] == 'R' && line[2] == 'R'
+	  && line[3] == ' ')
+	err = map_assuan_error (atoi (&line[4]));
+      else if (linelen >= 2
+	       && line[0] == 'S' && line[1] == ' ')
+	{
+	  char *rest;
+	  GpgStatusCode r;
+
+	  rest = strchr (line + 2, ' ');
+	  if (!rest)
+	    rest = line + linelen; /* set to an empty string */
+	  else
+	    *(rest++) = 0;
+
+	  r = parse_status (line + 2);
+
+	  if (r >= 0 && status_fnc)
+	    status_fnc (status_fnc_value, r, rest);
+	  else
+	    err = mk_error (General_Error);
+	}
+      else
+	err = mk_error (General_Error);
     }
-  while (*line == '#' || !linelen);
-  
-  if (linelen >= 2
-      && line[0] == 'O' && line[1] == 'K'
-      && (line[2] == '\0' || line[2] == ' '))
-    return 0;
+  while (!err);
 
-  if (linelen >= 4
-      && line[0] == 'E' && line[1] == 'R' && line[2] == 'R'
-      && line[3] == ' ')
-    err = map_assuan_error (atoi (&line[4]));
-
-  if (!err)
-    err = mk_error (General_Error);
-  return 0;
+  return err;
 }
 
 
@@ -384,7 +613,27 @@ gpgsm_set_fd (ASSUAN_CONTEXT ctx, const char *which, int fd, const char *opt)
   else
     snprintf (line, COMMANDLINELEN, "%s FD=%i", which, fd);
 
-  return gpgsm_assuan_simple_command (ctx, line);
+  return gpgsm_assuan_simple_command (ctx, line, NULL, NULL);
+}
+
+
+static const char *
+map_input_enc (GpgmeData d)
+{
+  switch (gpgme_data_get_encoding (d))
+    {
+    case GPGME_DATA_ENCODING_NONE:
+      break;
+    case GPGME_DATA_ENCODING_BINARY:
+      return "--binary";
+    case GPGME_DATA_ENCODING_BASE64:
+      return "--base64";
+    case GPGME_DATA_ENCODING_ARMOR:
+      return "--armor";
+    default:
+      break;
+    }
+  return NULL;
 }
 
 
@@ -400,15 +649,16 @@ _gpgme_gpgsm_op_decrypt (GpgsmObject gpgsm, GpgmeData ciph, GpgmeData plain)
   if (!gpgsm->command)
     return mk_error (Out_Of_Core);
 
-  gpgsm->input_data = ciph;
-  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server, 0);
+  gpgsm->input_cb.data = ciph;
+  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server, 
+                      map_input_enc (gpgsm->input_cb.data));
   if (err)
     return mk_error (General_Error);	/* FIXME */
-  gpgsm->output_data = plain;
+  gpgsm->output_cb.data = plain;
   err = gpgsm_set_fd (gpgsm->assuan_ctx, "OUTPUT", gpgsm->output_fd_server, 0);
   if (err)
     return mk_error (General_Error);	/* FIXME */
-  _gpgme_io_close (gpgsm->message_fd);
+  _gpgme_io_close (gpgsm->message_cb.fd);
 
   return 0;
 }
@@ -446,23 +696,17 @@ gpgsm_set_recipients (GpgsmObject gpgsm, GpgmeRecipients recp)
 	  if (! newline)
 	    {
 	      xfree (line);
-	      return mk_error(Out_Of_Core);
+	      return mk_error (Out_Of_Core);
 	    }
 	  line = newline;
 	  linelen = newlen;
 	}
       strcpy (&line[10], r->name);
       
-      err = gpgsm_assuan_simple_command (ctx, line);
+      err = gpgsm_assuan_simple_command (ctx, line, gpgsm->status.fnc,
+					 gpgsm->status.fnc_value);
       if (!err)
 	valid_recipients = 1;
-      else if (err == GPGME_Invalid_Key && gpgsm->status.fnc)
-	{
-	  /* FIXME: Include other reasons.  */
-	  line[8] = '0';	/* FIXME: Report detailed reason.  */
-	  gpgsm->status.fnc (gpgsm->status.fnc_value, STATUS_INV_RECP, &line[8]);
-	  line[8] = 'T';
-	}
       else if (err != GPGME_Invalid_Key)
 	{
 	  xfree (line);
@@ -491,16 +735,17 @@ _gpgme_gpgsm_op_encrypt (GpgsmObject gpgsm, GpgmeRecipients recp,
   if (!gpgsm->command)
     return mk_error (Out_Of_Core);
 
-  gpgsm->input_data = plain;
-  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server, 0);
+  gpgsm->input_cb.data = plain;
+  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server,
+                      map_input_enc (gpgsm->input_cb.data));
   if (err)
     return err;
-  gpgsm->output_data = ciph;
+  gpgsm->output_cb.data = ciph;
   err = gpgsm_set_fd (gpgsm->assuan_ctx, "OUTPUT", gpgsm->output_fd_server,
 		      use_armor ? "--armor" : 0);
   if (err)
     return err;
-  _gpgme_io_close (gpgsm->message_fd);
+  _gpgme_io_close (gpgsm->message_cb.fd);
 
   err = gpgsm_set_recipients (gpgsm, recp);
   if (err)
@@ -514,8 +759,62 @@ GpgmeError
 _gpgme_gpgsm_op_export (GpgsmObject gpgsm, GpgmeRecipients recp,
 			GpgmeData keydata, int use_armor)
 {
-  /* FIXME */
-  return mk_error (Not_Implemented);
+  GpgmeError err = 0;
+  char *cmd = NULL;
+  int cmdi;
+  int cmdlen = 32;
+
+  if (!gpgsm)
+    return mk_error (Invalid_Value);
+
+  cmd = malloc (cmdlen);
+  if (!cmd)
+    return mk_error (Out_Of_Core);
+  strcpy (cmd, "EXPORT");
+  cmdi = 6;
+
+  if (recp)
+    {
+      void *ec;
+      const char *s;
+
+      err = gpgme_recipients_enum_open (recp, &ec);
+      while (!err && (s = gpgme_recipients_enum_read (recp, &ec)))
+	{
+	  int slen = strlen (s);
+	  /* New string is old string + ' ' + s + '\0'.  */
+	  if (cmdlen < cmdi + 1 + slen + 1)
+	    {
+	      char *newcmd = xtryrealloc (cmd, cmdlen * 2);
+	      if (!newcmd)
+		{
+		  xfree (cmd);
+		  return mk_error (Out_Of_Core);
+		}
+	      cmd = newcmd;
+	      cmdlen *= 2;
+	    }
+	  cmd[cmdi++] = ' ';
+	  strcpy (cmd + cmdi, s);
+	  cmdi += slen;
+	}
+      if (!err)
+	err = gpgme_recipients_enum_close (recp, &ec);
+      if (err)
+	return err;
+    }
+
+  gpgsm->command = cmd;
+
+  gpgsm->output_cb.data = keydata;
+  err = gpgsm_set_fd (gpgsm->assuan_ctx, "OUTPUT", gpgsm->output_fd_server,
+		      use_armor ? "--armor" : 0);
+  if (err)
+    return err;
+  _gpgme_io_close (gpgsm->input_cb.fd);
+  _gpgme_io_close (gpgsm->message_cb.fd);
+
+  return 0;
 }
 
 
@@ -532,16 +831,17 @@ _gpgme_gpgsm_op_genkey (GpgsmObject gpgsm, GpgmeData help_data, int use_armor,
   if (!gpgsm->command)
     return mk_error (Out_Of_Core);
 
-  gpgsm->input_data = help_data;
-  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server, 0);
+  gpgsm->input_cb.data = help_data;
+  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server,
+                      map_input_enc (gpgsm->input_cb.data));
   if (err)
     return err;
-  gpgsm->output_data = pubkey;
+  gpgsm->output_cb.data = pubkey;
   err = gpgsm_set_fd (gpgsm->assuan_ctx, "OUTPUT", gpgsm->output_fd_server,
 		      use_armor ? "--armor" : 0);
   if (err)
     return err;
-  _gpgme_io_close (gpgsm->message_fd);
+  _gpgme_io_close (gpgsm->message_cb.fd);
 
   return 0;
 }
@@ -559,12 +859,13 @@ _gpgme_gpgsm_op_import (GpgsmObject gpgsm, GpgmeData keydata)
   if (!gpgsm->command)
     return mk_error (Out_Of_Core);
 
-  gpgsm->input_data = keydata;
-  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server, 0);
+  gpgsm->input_cb.data = keydata;
+  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server,
+                      map_input_enc (gpgsm->input_cb.data));
   if (err)
     return err;
-  _gpgme_io_close (gpgsm->output_fd);
-  _gpgme_io_close (gpgsm->message_fd);
+  _gpgme_io_close (gpgsm->output_cb.fd);
+  _gpgme_io_close (gpgsm->message_cb.fd);
 
   return 0;
 }
@@ -575,10 +876,18 @@ _gpgme_gpgsm_op_keylist (GpgsmObject gpgsm, const char *pattern,
 			 int secret_only, int keylist_mode)
 {
   char *line;
+  GpgmeError err;
 
   if (!pattern)
     pattern = "";
 
+  if (asprintf (&line, "OPTION list-mode=%d", (keylist_mode & 3)) < 0)
+    return mk_error (Out_Of_Core);
+  err = gpgsm_assuan_simple_command (gpgsm->assuan_ctx, line, NULL, NULL);
+  free (line);
+  if (err)
+    return err;
+
   /* Length is "LISTSECRETKEYS " + p + '\0'.  */
   line = xtrymalloc (15 + strlen (pattern) + 1);
   if (!line)
@@ -594,9 +903,9 @@ _gpgme_gpgsm_op_keylist (GpgsmObject gpgsm, const char *pattern,
       strcpy (&line[9], pattern);
     }
 
-  _gpgme_io_close (gpgsm->input_fd);
-  _gpgme_io_close (gpgsm->output_fd);
-  _gpgme_io_close (gpgsm->message_fd);
+  _gpgme_io_close (gpgsm->input_cb.fd);
+  _gpgme_io_close (gpgsm->output_cb.fd);
+  _gpgme_io_close (gpgsm->message_cb.fd);
 
   gpgsm->command = line;
   return 0;
@@ -608,13 +917,21 @@ _gpgme_gpgsm_op_keylist_ext (GpgsmObject gpgsm, const char *pattern[],
 			     int secret_only, int reserved, int keylist_mode)
 {
   char *line;
+  GpgmeError err;
   /* Length is "LISTSECRETKEYS " + p + '\0'.  */
   int length = 15 + 1;
   char *linep;
-
+  
   if (reserved)
     return mk_error (Invalid_Value);
 
+  if (asprintf (&line, "OPTION list-mode=%d", (keylist_mode & 3)) < 0)
+    return mk_error (Out_Of_Core);
+  err = gpgsm_assuan_simple_command (gpgsm->assuan_ctx, line, NULL, NULL);
+  free (line);
+  if (err)
+    return err;
+
   if (pattern && *pattern)
     {
       const char **pat = pattern;
@@ -685,9 +1002,9 @@ _gpgme_gpgsm_op_keylist_ext (GpgsmObject gpgsm, const char *pattern[],
     }
   *linep = '\0';
 
-  _gpgme_io_close (gpgsm->input_fd);
-  _gpgme_io_close (gpgsm->output_fd);
-  _gpgme_io_close (gpgsm->message_fd);
+  _gpgme_io_close (gpgsm->input_cb.fd);
+  _gpgme_io_close (gpgsm->output_cb.fd);
+  _gpgme_io_close (gpgsm->message_cb.fd);
 
   gpgsm->command = line;
   return 0;
@@ -713,21 +1030,22 @@ _gpgme_gpgsm_op_sign (GpgsmObject gpgsm, GpgmeData in, GpgmeData out,
 
   if (asprintf (&assuan_cmd, "OPTION include-certs %i", include_certs) < 0)
     return mk_error (Out_Of_Core);
-  err = gpgsm_assuan_simple_command (gpgsm->assuan_ctx, assuan_cmd);
+  err = gpgsm_assuan_simple_command (gpgsm->assuan_ctx, assuan_cmd, NULL, NULL);
   free (assuan_cmd);
   if (err)
     return err;
 
-  gpgsm->input_data = in;
-  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server, 0);
+  gpgsm->input_cb.data = in;
+  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server,
+                      map_input_enc (gpgsm->input_cb.data));
   if (err)
     return err;
-  gpgsm->output_data = out;
+  gpgsm->output_cb.data = out;
   err = gpgsm_set_fd (gpgsm->assuan_ctx, "OUTPUT", gpgsm->output_fd_server,
 		      use_armor ? "--armor" : 0);
   if (err)
     return err;
-  _gpgme_io_close (gpgsm->message_fd);
+  _gpgme_io_close (gpgsm->message_cb.fd);
 
   return 0;
 }
@@ -753,27 +1071,29 @@ _gpgme_gpgsm_op_verify (GpgsmObject gpgsm, GpgmeData sig, GpgmeData text)
   if (!gpgsm->command)
     return mk_error (Out_Of_Core);
 
-  gpgsm->input_data = sig;
-  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server, 0);
+  gpgsm->input_cb.data = sig;
+  err = gpgsm_set_fd (gpgsm->assuan_ctx, "INPUT", gpgsm->input_fd_server,
+                      map_input_enc (gpgsm->input_cb.data));
   if (err)
     return err;
   if (_gpgme_data_get_mode (text) == GPGME_DATA_MODE_IN)
     {
       /* Normal or cleartext signature.  */
-      gpgsm->output_data = text;
+      gpgsm->output_cb.data = text;
       err = gpgsm_set_fd (gpgsm->assuan_ctx, "OUTPUT", gpgsm->output_fd_server,
 			  0);
+      _gpgme_io_close (gpgsm->message_cb.fd);
     }
   else
     {
       /* Detached signature.  */
-      gpgsm->message_data = text;
+      gpgsm->message_cb.data = text;
       err = gpgsm_set_fd (gpgsm->assuan_ctx, "MESSAGE",
 			  gpgsm->message_fd_server, 0);
+      _gpgme_io_close (gpgsm->output_cb.fd);
     }
   if (err)
     return err;
-  _gpgme_io_close (gpgsm->output_fd);
 
   return 0;
 }
@@ -789,8 +1109,19 @@ status_cmp (const void *ap, const void *bp)
 }
 
 
-static int
-gpgsm_status_handler (void *opaque, int pid, int fd)
+static GpgStatusCode
+parse_status (const char *name)
+{
+  struct status_table_s t, *r;
+  t.name = name;
+  r = bsearch (&t, status_table, DIM(status_table) - 1,
+	       sizeof t, status_cmp);
+  return r ? r->code : -1;
+}
+
+
+static void
+gpgsm_status_handler (void *opaque, int fd)
 {
   AssuanError err;
   GpgsmObject gpgsm = opaque;
@@ -809,8 +1140,9 @@ gpgsm_status_handler (void *opaque, int pid, int fd)
 	      && line[0] == 'E' && line[1] == 'R' && line[2] == 'R'
 	      && (line[3] == '\0' || line[3] == ' ')))
 	{
-	  /* XXX: If an error occured, find out what happened, then save the error value
-	     before running the status handler (so it takes precedence).  */
+	  /* XXX: If an error occured, find out what happened, then
+	     save the error value before running the status handler
+	     (so it takes precedence).  */
 	  if (!err && line[0] == 'E' && line[3] == ' ')
 	    {
 	      err = map_assuan_error (atoi (&line[4]));
@@ -833,7 +1165,8 @@ gpgsm_status_handler (void *opaque, int pid, int fd)
 	  if (err)
 	    assuan_write_line (gpgsm->assuan_ctx, "BYE");
 
-	  return 1;
+	  _gpgme_io_close (gpgsm->status_cb.fd);
+	  return;
 	}
 
       if (linelen > 2
@@ -858,7 +1191,10 @@ gpgsm_status_handler (void *opaque, int pid, int fd)
 	      unsigned char *newline = xtryrealloc (*aline,
 						    *alinelen + linelen + 1);
 	      if (!newline)
-		return mk_error (Out_Of_Core);
+		{
+		  _gpgme_io_close (gpgsm->status_cb.fd);
+		  return;
+		}
 	      *aline = newline;
 	      gpgsm->colon.attic.linesize += linelen + 1;
 	    }
@@ -902,31 +1238,27 @@ gpgsm_status_handler (void *opaque, int pid, int fd)
       else if (linelen > 2
 	  && line[0] == 'S' && line[1] == ' ')
 	{
-	  struct status_table_s t, *r;
 	  char *rest;
-	  
+	  GpgStatusCode r;
+
 	  rest = strchr (line + 2, ' ');
 	  if (!rest)
 	    rest = line + linelen; /* set to an empty string */
 	  else
-	    *rest++ = 0;
+	    *(rest++) = 0;
 
-	  t.name = line + 2;
-	  r = bsearch (&t, status_table, DIM(status_table) - 1,
-		       sizeof t, status_cmp);
+	  r = parse_status (line + 2);
 
-	  if (r)
+	  if (r >= 0)
 	    {
 	      if (gpgsm->status.fnc)
-		gpgsm->status.fnc (gpgsm->status.fnc_value, r->code, rest);
+		gpgsm->status.fnc (gpgsm->status.fnc_value, r, rest);
 	    }
 	  else
-	    fprintf (stderr, "[UNKNOWN STATUS]%s %s", t.name, rest);
+	    fprintf (stderr, "[UNKNOWN STATUS]%s %s", line + 2, rest);
 	}
     }
   while (assuan_pending_line (gpgsm->assuan_ctx));
-  
-  return 0;
 }
 
 
@@ -952,50 +1284,45 @@ _gpgme_gpgsm_set_colon_line_handler (GpgsmObject gpgsm,
 }
 
 
+static GpgmeError
+_gpgme_gpgsm_add_io_cb (GpgsmObject gpgsm, iocb_data_t *iocbd,
+			GpgmeIOCb handler)
+{
+  GpgmeError err = 0;
+
+  iocbd->tag = (*gpgsm->io_cbs.add) (gpgsm->io_cbs.add_priv,
+				     iocbd->fd, iocbd->dir,
+				     handler, iocbd->data);
+  if (!iocbd->tag)
+    err = mk_error (General_Error);
+  if (!err && !iocbd->dir)
+    /* FIXME Kludge around poll() problem.  */
+    err = _gpgme_io_set_nonblocking (iocbd->fd);
+  return err;
+}
+
 GpgmeError
 _gpgme_gpgsm_start (GpgsmObject gpgsm, void *opaque)
 {
   GpgmeError err = 0;
   pid_t pid;
-  int fdlist[5];
-  int nfds;
 
   if (!gpgsm)
     return mk_error (Invalid_Value);
 
   pid = assuan_get_pid (gpgsm->assuan_ctx);
 
-  /* We need to know the fd used by assuan for reads.  We do this by
-     using the assumption that the first returned fd from
-     assuan_get_active_fds() is always this one. */
-  nfds = assuan_get_active_fds (gpgsm->assuan_ctx, 0 /* read fds */,
-                                fdlist, DIM (fdlist));
-  if (nfds < 1)
-    return mk_error (General_Error);  /* FIXME */
-  err = _gpgme_register_pipe_handler (opaque, gpgsm_status_handler, gpgsm, pid,
-                                      fdlist[0], 1);
-
-
-  if (gpgsm->input_fd != -1)
-    {
-      err = _gpgme_register_pipe_handler (opaque, _gpgme_data_outbound_handler,
-					  gpgsm->input_data, pid,
-					  gpgsm->input_fd, 0);
-      if (!err)	/* FIXME Kludge around poll() problem.  */
-	err = _gpgme_io_set_nonblocking (gpgsm->input_fd);
-    }
-  if (!err && gpgsm->output_fd != -1)
-    err = _gpgme_register_pipe_handler (opaque, _gpgme_data_inbound_handler,
-					gpgsm->output_data, pid,
-					gpgsm->output_fd, 1);
-  if (!err && gpgsm->message_fd != -1)
-    {
-      err = _gpgme_register_pipe_handler (opaque, _gpgme_data_outbound_handler,
-					  gpgsm->message_data, pid,
-					  gpgsm->message_fd, 0);
-      if (!err)	/* FIXME Kludge around poll() problem.  */
-	err = _gpgme_io_set_nonblocking (gpgsm->message_fd);
-    }
+  err = _gpgme_gpgsm_add_io_cb (gpgsm, &gpgsm->status_cb,
+				gpgsm_status_handler);
+  if (gpgsm->input_cb.fd != -1)
+    err = _gpgme_gpgsm_add_io_cb (gpgsm, &gpgsm->input_cb,
+				  _gpgme_data_outbound_handler);
+  if (!err && gpgsm->output_cb.fd != -1)
+    err = _gpgme_gpgsm_add_io_cb (gpgsm, &gpgsm->output_cb,
+				  _gpgme_data_inbound_handler);
+  if (!err && gpgsm->message_cb.fd != -1)
+    err = _gpgme_gpgsm_add_io_cb (gpgsm, &gpgsm->message_cb,
+				  _gpgme_data_outbound_handler);
 
   if (!err)
     err = assuan_write_line (gpgsm->assuan_ctx, gpgsm->command);
@@ -1003,6 +1330,12 @@ _gpgme_gpgsm_start (GpgsmObject gpgsm, void *opaque)
   return err;
 }
 
+void
+_gpgme_gpgsm_set_io_cbs (GpgsmObject gpgsm, struct GpgmeIOCbs *io_cbs)
+{
+  gpgsm->io_cbs = *io_cbs;
+}
+
 
 #else	/* ENABLE_GPGSM */
 
@@ -1146,5 +1479,9 @@ _gpgme_gpgsm_start (GpgsmObject gpgsm, void *opaque)
   return mk_error (Invalid_Engine);
 }
 
+void
+_gpgme_gpgsm_set_io_cbs (GpgsmObject gpgsm, struct GpgmeIOCbs *io_cbs)
+{
+}
 
 #endif	/* ! ENABLE_GPGSM */
