@@ -28,7 +28,7 @@
 #include "util.h"
 #include "context.h"
 #include "ops.h"
-
+#include "wait.h"
 
 /**
  * gpgme_new:
@@ -53,8 +53,8 @@ gpgme_new (GpgmeCtx *r_ctx)
   ctx->keylist_mode = GPGME_KEYLIST_MODE_LOCAL;
   ctx->verbosity = 1;
   ctx->include_certs = 1;
+  _gpgme_fd_table_init (&ctx->fdt);
   *r_ctx = ctx;
-
   return 0;
 }
 
@@ -71,6 +71,7 @@ gpgme_release (GpgmeCtx ctx)
   if (!ctx)
     return;
   _gpgme_engine_release (ctx->engine);
+  _gpgme_fd_table_deinit (&ctx->fdt);
   _gpgme_release_result (ctx);
   gpgme_key_release (ctx->tmp_key);
   gpgme_data_release (ctx->help_data_1);
@@ -93,6 +94,7 @@ _gpgme_release_result (GpgmeCtx ctx)
   _gpgme_release_import_result (ctx->result.import);
   _gpgme_release_delete_result (ctx->result.delete);
   _gpgme_release_genkey_result (ctx->result.genkey);
+  _gpgme_release_keylist_result (ctx->result.keylist);
   memset (&ctx->result, 0, sizeof (ctx->result));
   _gpgme_set_op_info (ctx, NULL);
   ctx->error = 0;
@@ -140,7 +142,7 @@ gpgme_get_notation (GpgmeCtx ctx)
  * @c: the context
  * @reserved:
  *
- * Return information about the last information.  The caller has to
+ * Return information about the last operation.  The caller has to
  * free the string.  NULL is returned if there is not previous
  * operation available or the operation has not yet finished.
  *
@@ -212,6 +214,16 @@ gpgme_set_protocol (GpgmeCtx ctx, GpgmeProtocol protocol)
   return 0;
 }
 
+GpgmeProtocol
+gpgme_get_protocol (GpgmeCtx ctx)
+{
+  if (!ctx)
+    return 0; /* well, this is OpenPGP */
+  if (ctx->use_cms)
+    return GPGME_PROTOCOL_CMS;
+  return GPGME_PROTOCOL_OpenPGP;
+}
+
 
 /**
  * gpgme_set_armor:
@@ -397,6 +409,36 @@ gpgme_set_passphrase_cb (GpgmeCtx ctx, GpgmePassphraseCb cb, void *cb_value)
     }
 }
 
+
+/**
+ * gpgme_get_passphrase_cb:
+ * @ctx: the context
+ * @r_cb: The current callback function
+ * @r_cb_value: The current value passed to the callback function
+ *
+ * This function returns the callback function to be used to pass a passphrase
+ * to the crypto engine.
+ **/
+void
+gpgme_get_passphrase_cb (GpgmeCtx ctx, GpgmePassphraseCb *r_cb, void **r_cb_value)
+{
+  if (ctx)
+    {
+      if (r_cb)
+	*r_cb = ctx->passphrase_cb;
+      if (r_cb_value)
+	*r_cb_value = ctx->passphrase_cb_value;
+    }
+  else
+    {
+      if (r_cb)
+	*r_cb = NULL;
+      if (r_cb_value)
+	*r_cb_value = NULL;
+    }
+}
+
+
 /**
  * gpgme_set_progress_cb:
  * @ctx: the context
@@ -423,3 +465,88 @@ gpgme_set_progress_cb (GpgmeCtx ctx, GpgmeProgressCb cb, void *cb_value)
       ctx->progress_cb_value = cb_value;
     }
 }
+
+
+/**
+ * gpgme_get_progress_cb:
+ * @ctx: the context
+ * @r_cb: The current callback function
+ * @r_cb_value: The current value passed to the callback function
+ *
+ * This function returns the callback function to be used as a
+ * progress indicator.
+ **/
+void
+gpgme_get_progress_cb (GpgmeCtx ctx, GpgmeProgressCb *r_cb, void **r_cb_value)
+{
+  if (ctx)
+    {
+      if (r_cb)
+	*r_cb = ctx->progress_cb;
+      if (r_cb_value)
+	*r_cb_value = ctx->progress_cb_value;
+    }
+  else
+    {
+      if (r_cb)
+	*r_cb = NULL;
+      if (r_cb_value)
+	*r_cb_value = NULL;
+    }
+}
+
+
+/**
+ * gpgme_set_io_cbs:
+ * @ctx: the context
+ * @register_io_cb: A callback function
+ * @register_hook_value: The value passed to the callback function
+ * @remove_io_cb: Another callback function
+ *
+ **/
+void
+gpgme_set_io_cbs (GpgmeCtx ctx, struct GpgmeIOCbs *io_cbs)
+{
+  if (ctx && io_cbs)
+    ctx->io_cbs = *io_cbs;
+  else
+    {
+      ctx->io_cbs.add = NULL;
+      ctx->io_cbs.add_priv = NULL;
+      ctx->io_cbs.remove = NULL;
+      ctx->io_cbs.event = NULL;
+      ctx->io_cbs.event_priv = NULL;
+    }
+}
+
+
+/**
+ * gpgme_get_io_cbs:
+ * @ctx: the context
+ * @r_register_cb: The current register callback function
+ * @r_register_cb_value: The current value passed to the
+ * register callback function
+ * @r_remove_cb: The current remove callback function
+ *
+ * This function returns the callback function to be used to pass a passphrase
+ * to the crypto engine.
+ **/
+void
+gpgme_get_io_cbs (GpgmeCtx ctx, struct GpgmeIOCbs *io_cbs)
+{
+  if (ctx && io_cbs)
+    *io_cbs = ctx->io_cbs;
+}
+
+
+void
+_gpgme_op_event_cb (void *data, GpgmeEventIO type, void *type_data)
+{
+  GpgmeCtx ctx = data;
+
+  if (type == GPGME_EVENT_DONE)
+    ctx->pending = 0;
+
+  if (ctx->io_cbs.add && ctx->io_cbs.event)
+    (*ctx->io_cbs.event) (ctx->io_cbs.event_priv, type, type_data);
+}
