@@ -33,19 +33,83 @@
 
 #define my_isdigit(a) ( (a) >='0' && (a) <= '9' )
 
+struct keylist_result_s
+{
+  int truncated;
+  GpgmeData xmlinfo;
+};
+
 static void finish_key ( GpgmeCtx ctx );
 
 
+void
+_gpgme_release_keylist_result (KeylistResult result)
+{
+  if (!result)
+    return;
+  xfree (result);
+}
+
+/* Append some XML info.  args is currently ignore but we might want
+   to add more information in the future (like source of the
+   keylisting.  With args of NULL the XML structure is closed.  */
+static void
+append_xml_keylistinfo (GpgmeData *rdh, char *args)
+{
+  GpgmeData dh;
+
+  if (!*rdh)
+    {
+      if (gpgme_data_new (rdh))
+	return; /* FIXME: We are ignoring out-of-core.  */
+      dh = *rdh;
+      _gpgme_data_append_string (dh, "<GnupgOperationInfo>\n");
+    }
+  else
+    {
+      dh = *rdh;
+      _gpgme_data_append_string (dh, "  </keylisting>\n");
+    }
+
+  if (!args)
+    {
+      /* Just close the XML containter.  */
+      _gpgme_data_append_string (dh, "</GnupgOperationInfo>\n");
+      return;
+    }
+
+  _gpgme_data_append_string (dh,
+                             "  <keylisting>\n"
+                             "    <truncated/>\n"
+			     );
+    
+}
+
+
+
 static void
 keylist_status_handler (GpgmeCtx ctx, GpgStatusCode code, char *args)
 {
   if (ctx->error)
     return;
+  test_and_allocate_result (ctx, keylist);
 
   switch (code)
     {
+    case STATUS_TRUNCATED:
+      ctx->result.keylist->truncated = 1;
+      break;
+
     case STATUS_EOF:
       finish_key (ctx);
+      if (ctx->result.keylist->truncated)
+        append_xml_keylistinfo (&ctx->result.keylist->xmlinfo, "1");
+      if (ctx->result.keylist->xmlinfo)
+	{
+	  append_xml_keylistinfo (&ctx->result.keylist->xmlinfo, NULL);
+	  _gpgme_set_op_info (ctx, ctx->result.keylist->xmlinfo);
+	  ctx->result.keylist->xmlinfo = NULL;
+        }
       break;
 
     default:
@@ -151,6 +215,23 @@ set_subkey_capability ( struct subkey_s *k, const char *s)
     }
 }
 
+static void
+set_ownertrust (GpgmeKey key, const char *s)
+{
+  /* Look at letters and stop at the first digit.  */
+  for (; *s && !my_isdigit (*s); s++)
+    {
+      switch (*s)
+	{
+	case 'n': key->otrust = GPGME_VALIDITY_NEVER; break;
+	case 'm': key->otrust = GPGME_VALIDITY_MARGINAL; break;
+	case 'f': key->otrust = GPGME_VALIDITY_FULL; break;
+	case 'u': key->otrust = GPGME_VALIDITY_ULTIMATE; break;
+        default : key->otrust = GPGME_VALIDITY_UNKNOWN; break;
+        }
+    }
+}
+
 
 /* Note: We are allowed to modify LINE.  */
 static void
@@ -219,47 +300,51 @@ keylist_colon_handler (GpgmeCtx ctx, char *line)
 	      /* Start a new keyblock.  */
 	      if (_gpgme_key_new (&key))
 		{
-		  ctx->error = mk_error (Out_Of_Core);  /* the only kind of error we can get*/
+		  /* The only kind of error we can get.  */
+		  ctx->error = mk_error (Out_Of_Core);
 		  return;
                 }
-                rectype = RT_PUB;
-                finish_key (ctx);
-                assert (!ctx->tmp_key);
-                ctx->tmp_key = key;
+	      rectype = RT_PUB;
+	      finish_key (ctx);
+	      assert (!ctx->tmp_key);
+	      ctx->tmp_key = key;
             }
-            else if (!strcmp (p, "sec"))
-	      {
-                /* Start a new keyblock,  */
-                if (_gpgme_key_new_secret (&key))
-		  {
-                    ctx->error = mk_error (Out_Of_Core);  /* The only kind of error we can get*/
-                    return;
-                }
-                rectype = RT_SEC;
-                finish_key (ctx);
-                assert (!ctx->tmp_key);
-                ctx->tmp_key = key;
+	  else if (!strcmp (p, "sec"))
+	    {
+	      /* Start a new keyblock,  */
+	      if (_gpgme_key_new_secret (&key))
+		{
+		  /* The only kind of error we can get.  */
+		  ctx->error = mk_error (Out_Of_Core);
+		  return;
+		}
+	      rectype = RT_SEC;
+	      finish_key (ctx);
+	      assert (!ctx->tmp_key);
+	      ctx->tmp_key = key;
             }
-            else if (!strcmp (p, "crt"))
-	      {
-                /* Start a new certificate. */
-                if (_gpgme_key_new (&key))
-		  {
-                    ctx->error = mk_error (Out_Of_Core);  /* The only kind of error we can get*/
-                    return;
-                }
-                key->x509 = 1;
-                rectype = RT_CRT;
-                finish_key (ctx);
-                assert (!ctx->tmp_key);
-                ctx->tmp_key = key;
+	  else if (!strcmp (p, "crt"))
+	    {
+	      /* Start a new certificate. */
+	      if (_gpgme_key_new (&key))
+		{
+		  /* The only kind of error we can get.  */
+		  ctx->error = mk_error (Out_Of_Core);
+		  return;
+		}
+	      key->x509 = 1;
+	      rectype = RT_CRT;
+	      finish_key (ctx);
+	      assert (!ctx->tmp_key);
+	      ctx->tmp_key = key;
             }
 	  else if (!strcmp (p, "crs"))
 	    {
 	      /* Start a new certificate.  */
 	      if (_gpgme_key_new_secret (&key))
 		{
-		  ctx->error = mk_error (Out_Of_Core);  /* The only kind of error we can get*/
+		  /* The only kind of error we can get.  */
+		  ctx->error = mk_error (Out_Of_Core);
 		  return;
                 }
 	      key->x509 = 1;
@@ -272,7 +357,6 @@ keylist_colon_handler (GpgmeCtx ctx, char *line)
 	    rectype = RT_FPR;
 	  else 
 	    rectype = RT_NONE;
-            
         }
       else if (rectype == RT_PUB || rectype == RT_SEC
 	       || rectype == RT_CRT || rectype == RT_CRS)
@@ -312,6 +396,7 @@ keylist_colon_handler (GpgmeCtx ctx, char *line)
                 }
 	      break;
 	    case 9: /* ownertrust */
+              set_ownertrust (key, p);
 	      break;
 	    case 10: /* not used for gpg due to --fixed-list-mode option 
 			but gpgsm stores the issuer name */
@@ -383,7 +468,8 @@ keylist_colon_handler (GpgmeCtx ctx, char *line)
 	      break;
 	    case 10: /* user ID */
 	      if (_gpgme_key_append_name (key, p))
-		ctx->error = mk_error (Out_Of_Core);  /* The only kind of error we can get*/
+		/* The only kind of error we can get*/
+		ctx->error = mk_error (Out_Of_Core);
 	      else
 		{
 		  if (trust_info)
@@ -477,25 +563,13 @@ gpgme_op_keylist_start (GpgmeCtx ctx, const char *pattern, int secret_only)
 {
   GpgmeError err = 0;
 
-  if (!ctx)
-    return mk_error (Invalid_Value);
-  ctx->pending = 1;
+  err = _gpgme_op_reset (ctx, 0);
+  if (err)
+    goto leave;
 
-  _gpgme_release_result (ctx);
-
-  if (ctx->engine)
-    {
-      _gpgme_engine_release (ctx->engine); 
-      ctx->engine = NULL;
-    }
   gpgme_key_release (ctx->tmp_key);
   ctx->tmp_key = NULL;
   /* Fixme: Release key_queue.  */
-    
-  err = _gpgme_engine_new (ctx->use_cms ? GPGME_PROTOCOL_CMS
-			   : GPGME_PROTOCOL_OpenPGP, &ctx->engine);
-  if (err)
-    goto leave;
 
   _gpgme_engine_set_status_handler (ctx->engine, keylist_status_handler, ctx);
   err = _gpgme_engine_set_colon_line_handler (ctx->engine,
@@ -508,7 +582,8 @@ gpgme_op_keylist_start (GpgmeCtx ctx, const char *pattern, int secret_only)
      just ignore those lines - This should speed up things */
   _gpgme_engine_set_verbosity (ctx->engine, 0);
 
-  err = _gpgme_engine_op_keylist (ctx->engine, pattern, secret_only, ctx->keylist_mode);
+  err = _gpgme_engine_op_keylist (ctx->engine, pattern, secret_only,
+				  ctx->keylist_mode);
 
   if (!err)	/* And kick off the process.  */
     err = _gpgme_engine_start (ctx->engine, ctx);
@@ -543,26 +618,12 @@ gpgme_op_keylist_ext_start (GpgmeCtx ctx, const char *pattern[],
 {
   GpgmeError err = 0;
 
-  if (!ctx)
-    return mk_error (Invalid_Value);
-  ctx->pending = 1;
-
-  _gpgme_release_result (ctx);
-
-  if (ctx->engine)
-    {
-      _gpgme_engine_release (ctx->engine); 
-      ctx->engine = NULL;
-    }
-  gpgme_key_release (ctx->tmp_key);
-  ctx->tmp_key = NULL;
-  /* Fixme: Release key_queue.  */
-    
-  err = _gpgme_engine_new (ctx->use_cms ? GPGME_PROTOCOL_CMS
-			   : GPGME_PROTOCOL_OpenPGP, &ctx->engine);
+  err = _gpgme_op_reset (ctx, 0);
   if (err)
     goto leave;
 
+  gpgme_key_release (ctx->tmp_key);
+
   _gpgme_engine_set_status_handler (ctx->engine, keylist_status_handler, ctx);
   err = _gpgme_engine_set_colon_line_handler (ctx->engine,
 					      keylist_colon_handler, ctx);
