@@ -1,6 +1,6 @@
 /* -*- Mode: C -*-
 
-  $Id: cryptplug.h,v 1.17 2002/03/22 10:40:24 khz Exp $
+  $Id: cryptplug.h,v 1.34 2002/06/25 17:42:21 steffen Exp $
 
   CRYPTPLUG - an independent cryptography plug-in API
 
@@ -23,6 +23,8 @@
 #ifndef CRYPTPLUG_H
 #define CRYPTPLUG_H
 
+#include <stdlib.h>
+
 #ifdef __cplusplus
 extern "C" {
 #else
@@ -31,11 +33,6 @@ typedef char bool;
 #define false 0
 #endif
 
-#include <stdlib.h>
-//#include <string.h>
-//#include <ctype.h>
-
-
 /*! \file cryptplug.h
     \brief Common API header for CRYPTPLUG.
 
@@ -160,6 +157,9 @@ typedef char bool;
     plugin <b>you should ignore this</b> section!
 */
 
+/*! \defgroup certList Certificate Info listing functions
+ */
+
 
 typedef enum {
   Feature_undef             = 0,
@@ -179,10 +179,11 @@ typedef enum {
   Feature_StoreMessagesEncrypted = 13,
   Feature_CheckCertificatePath = 14,
   Feature_CertificateDirectoryService = 15,
-  Feature_CRLDirectoryService = 16
+  Feature_CRLDirectoryService = 16,
+  Feature_CertificateInfo     = 17
 } Feature;
 
-// dummy values
+/* dummy values */
 typedef enum {
   PinRequest_undef            = 0,
 
@@ -193,7 +194,15 @@ typedef enum {
   PinRequest_AfterMinutes     = 5
 } PinRequests;
 
-// dummy values:
+
+typedef enum {
+  SignatureCompoundMode_undef    = 0,
+
+  SignatureCompoundMode_Opaque   = 1,
+  SignatureCompoundMode_Detached = 2
+} SignatureCompoundMode;
+
+
 typedef enum {
   SendCert_undef              = 0,
 
@@ -203,7 +212,7 @@ typedef enum {
   SendCert_SendChainWithRoot  = 4
 } SendCertificates;
 
-// dummy values:
+
 typedef enum {
   SignAlg_undef               = 0,
 
@@ -245,10 +254,65 @@ typedef enum {
 } CertificateSource;
 
 
+/*! \ingroup groupSignAct
+    \brief Flags used to compose the SigStatusFlags value.
+
+    This status flags are used to compose the SigStatusFlags value
+    returned in \c SignatureMetaDataExtendedInfo after trying to
+    verify a signed message part's signature status.
+
+    The normal flags may <b>not</b> be used together with the
+    special SigStat_NUMERICAL_CODE flag. When finding the special
+    SigStat_NUMERICAL_CODE flag in a SigStatusFlags value you
+    can obtain the respective error code number by substracting
+    the SigStatusFlags value by SigStat_NUMERICAL_CODE: this is
+    used to transport special status information NOT matching
+    any of the normal predefined status codes.
+
+    \note to PlugIn developers: Implementations of the CryptPlug API
+    should try to express their signature states by bit-wise OR'ing
+    the normal SigStatusFlags values. Using the SigStat_NUMERICAL_CODE
+    flag should only be used as for exceptional situations where no
+    other flag(s) could be used. By using the normal status flags your
+    PlugIn's users will be told an understandable description of the
+    status - when using (SigStat_NUMERICAL_CODE + internalCode) they
+    will only be shown the respective code number and have to look
+    into your PlugIn's manual to learn about it's meaning...
+*/
+enum {
+    SigStat_VALID       = 0x0001,   /* The signature is fully valid */
+    SigStat_GREEN       = 0x0002,   /* The signature is good. */
+    SigStat_RED         = 0x0004,   /* The signature is bad. */
+    SigStat_KEY_REVOKED = 0x0010,   /* One key has been revoked. */
+    SigStat_KEY_EXPIRED = 0x0020,   /* One key has expired. */
+    SigStat_SIG_EXPIRED = 0x0040,   /* The signature has expired. */
+    SigStat_KEY_MISSING = 0x0080,   /* Can't verify: key missing. */
+    SigStat_CRL_MISSING = 0x0100,   /* CRL not available. */
+    SigStat_CRL_TOO_OLD = 0x0200,   /* Available CRL is too old. */
+    SigStat_BAD_POLICY  = 0x0400,   /* A policy was not met. */
+    SigStat_SYS_ERROR   = 0x0800,   /* A system error occured. */
+
+    SigStat_NUMERICAL_CODE = 0x8000 /* An other error occured. */
+};
+typedef unsigned long SigStatusFlags;
 
 
 
 
+
+/*! \ingroup groupGeneral
+    \brief This function returns the version string of this cryptography
+           plug-in.
+
+   If the plugins initialization fails the calling process might want
+   to display the library version number to the user for checking if
+   there is an old version of the library installed...
+   
+   \note This function <b>must</b> be implemented by each plug-in using
+   this API specification.
+*/
+const char* libVersion( void );
+
 /*! \ingroup groupGeneral
     \brief This function returns a URL to be used for reporting a bug that
            you found (or suspect, resp.) in this cryptography plug-in.
@@ -405,6 +469,20 @@ void setSignatureAlgorithm( SignatureAlgorithm );
    \brief Returns the algorithm used for signing.
 */
 SignatureAlgorithm signatureAlgorithm( void );
+        
+/*! \ingroup groupConfigSign
+\brief Sets whether signatures and signed data should be send
+            as opaque signed or
+            as multipart/signed message parts.
+*/
+void setSignatureCompoundMode( SignatureCompoundMode );
+
+/*! \ingroup groupConfigSign
+\brief Returns whether signatures and signed data will be send
+            as opaque signed or
+            as multipart/signed message parts.
+*/
+SignatureCompoundMode signatureCompoundMode( void );
 
 /*! \ingroup groupConfigSign
    \brief Sets which certificates should be sent with the
@@ -1212,7 +1290,7 @@ struct StructuringInfo {
                                   FALSE) */
   bool  makeMimeObject;      /*!< specifies whether we should create a MIME
                                   object or a flat text message body */
-  // the following are used for MIME messages only
+  /* the following are used for MIME messages only */
   bool  makeMultiMime;       /*!< specifies whether we should create a
                                   'Multipart' MIME object or a single part
                                   object, if FALSE only \c contentTypeMain,
@@ -1279,7 +1357,7 @@ struct StructuringInfo {
                                   \c makeMimeObject or \c makeMultiMime
                                   is FALSE or if \c contentTypeCode does
                                   not return a non-zero-length string) */
-  // the following are used for flat non-MIME messages only
+  /* the following are used for flat non-MIME messages only */
   char* flatTextPrefix;      /*!< text to preceed the main text (or the
                                   code bloc containing the encrypted main
                                   text, resp.)<br>
@@ -1323,7 +1401,7 @@ struct StructuringInfo {
     \see free_StructuringInfo, StructuringInfo
     \see signMessage, encryptMessage, encryptAndSignMessage
 */
-  void init_StructuringInfo( struct StructuringInfo* s )
+  inline void init_StructuringInfo( struct StructuringInfo* s )
   {
     if( ! s ) return;
 
@@ -1364,7 +1442,7 @@ struct StructuringInfo {
 
     \see StructuringInfo
 */
-  void free_StructuringInfo( struct StructuringInfo* s )
+  inline void free_StructuringInfo( struct StructuringInfo* s )
   {
     if( ! s ) return;
     if( s->contentTypeMain )    free( s->contentTypeMain );
@@ -1420,10 +1498,12 @@ struct StructuringInfo {
    \see StructuringInfo, free_StructuringInfo
 */
 bool signMessage( const char*  cleartext,
-                  const char** ciphertext,
+                  char** ciphertext,
                   const size_t* cipherLen,
                   const char*  certificate,
-                  struct StructuringInfo* structuring );
+                  struct StructuringInfo* structuring,
+                  int* errId,
+                  char** errTxt );
 
 
 /*! \ingroup groupSignAct
@@ -1431,8 +1511,21 @@ bool signMessage( const char*  cleartext,
 struct SignatureMetaDataExtendedInfo
 {
     struct tm* creation_time;
+    SigStatusFlags sigStatusFlags;
     char* status_text;
+    char* keyid;
     char* fingerprint;
+    char* algo;
+    char* userid;
+    char* name;
+    char* email;
+    char* comment;
+    unsigned long algo_num;
+    unsigned long validity;
+    unsigned long userid_num;
+    unsigned long keylen;
+    unsigned long key_created;
+    unsigned long key_expires;
 };
 
 /*! \ingroup groupSignAct
@@ -1447,9 +1540,18 @@ struct SignatureMetaData {
 
 /*! \ingroup groupSignAct
    \brief Checks whether the signature of a message is
-          valid. \c ciphertext specifies the signed message
-          as it was received by the MUA, \c signaturetext is the
-          signature itself.
+          valid.
+
+   \c cleartext must never be 0 but be a valid pointer.
+
+   If \c *cleartext > 0 then **cleartext specifies the message text
+   that was signed and \c signaturetext is the signature itself.
+
+   If \c *cleartext == 0 is an empty string then \c signaturetext is
+   supposed to contain an opaque signed message part. After checking the
+   data and verifying the signature the cleartext of the message will be
+   returned in \c cleartext.  The user must free the respective memory
+   ocupied by *cleartext.
 
    Depending on the configuration, MUAs might not need to use this.
    If \c sigmeta is non-null, the
@@ -1457,7 +1559,7 @@ struct SignatureMetaData {
           contain meta information about the signature after the
           function call.
 */
-bool checkMessageSignature( const char* ciphertext,
+bool checkMessageSignature( char** cleartext,
                             const char* signaturetext,
                             bool signatureIsBinary,
                             int signatureLen,
@@ -1475,8 +1577,9 @@ bool storeCertificatesFromMessage( const char* ciphertext );
 
    NOTE: The \c certificate parameter must point to an allready allocated
    block of memory which is large enough to hold the complete list.
+   If secretOnly is true, only secret keys are returned.
 */
-bool findCertificates( const char* addressee, char** certificates );
+bool findCertificates( const char* addressee, char** certificates, bool secretOnly );
 
 /*! \ingroup groupCryptAct
    \brief Encrypts an email message in
@@ -1515,7 +1618,9 @@ bool encryptMessage( const char*  cleartext,
                      const char** ciphertext,
                      const size_t* cipherLen,
                      const char*  addressee,
-                     struct StructuringInfo* structuring );
+                     struct StructuringInfo* structuring,
+                     int* errId,
+                     char** errTxt );
 
 
 /*! \ingroup groupCryptAct
@@ -1596,12 +1701,11 @@ const char* requestCertificateDialog( void );
 
 /*! \ingroup groupCertAct
    \brief Generates a prototype certificate with the data provided
-        in the first four parameters and sends it via email to the CA
-          specified in \c ca_address.
+        in the four parameter. The memory returned in \a generatedKey
+        must be freed with free() by the caller.
 */
-bool requestDecentralCertificate( const char* name, const char*
-          email, const char* organization, const char* department,
-          const char* ca_address );
+bool requestDecentralCertificate( const char* certparms, 
+                                  char** generatedKey, int* keyLength );
 
 /*! \ingroup groupCertAct
    \brief Requests a certificate in a PSE from the CA
@@ -1679,6 +1783,70 @@ const char* displayCRL( void );
 */
 void updateCRL( void );
 
+struct CertIterator;
+
+struct DnPair {
+    char *key;
+    char *value;
+};
+
+struct CertificateInfo {
+  char** userid;
+  char* serial;
+  char* fingerprint;
+
+  char* issuer;
+  char* chainid;
+
+  char* caps;
+
+  unsigned long created;
+  unsigned long expire;
+
+  int secret   : 1;
+  int invalid  : 1;
+  int expired  : 1;
+  int disabled : 1;
+  
+
+  struct DnPair *dnarray; /* parsed values from userid[0] */
+};
+
+/*! \function struct CertIterator*  startListCertificates( const char* pattern );
+    \function struct CertificateInfo*  nextCertificate( struct CertIterator* );
+    \function void endListCertificates( struct CertIterator* );
+
+    \ingroup certList
+  Example that runs through certs matching "Steffen":
+\verbatim
+  struct CertificateInfo* info;
+  struct CertIterator* it = startListCertificates("Steffen", 0 );
+  while( nextCertificate( it, &info ) == GPGME_No_Error && info ) {
+    do something with info.
+    dont free() it, the struct will be reused
+    by the next call to nextCertificate()
+  }
+  endListCertificates( it );
+\endverbatim
+*/
+struct CertIterator*
+startListCertificates( const char* pattern, int remote );
+
+int
+nextCertificate( struct CertIterator*, struct CertificateInfo** result );
+
+void
+endListCertificates( struct CertIterator* );
+
+/*!
+  Import a certificate that was a result from a search-operation using the startListCertificates(), nextCertificate() funtions.
+
+  The fingerprint must be passed to identify the key.
+ */
+
+int
+importCertificate( const char* fingerprint );
+
 #ifdef __cplusplus
 }
 #endif
