@@ -37,6 +37,11 @@ char *make_dirname(const char *filepath);
 char *make_filename( const char *first_part, ... );
 int compare_filenames( const char *a, const char *b );
 
+size_t print_sanitized_buffer (FILE *fp, const void *buffer, size_t length,
+                               int delim);
+size_t print_sanitized_string (FILE *fp, const char *string, int delim);
+
+
 const char *ascii_memistr( const char *buf, size_t buflen, const char *sub );
 int ascii_isupper (int c);
 int ascii_islower (int c);
@@ -44,6 +49,8 @@ int ascii_toupper (int c);
 int ascii_tolower (int c);
 int ascii_strcasecmp( const char *a, const char *b );
 int ascii_memcasecmp( const char *a, const char *b, size_t n );
+void *ascii_memcasemem (const void *haystack, size_t nhaystack,
+                        const void *needle, size_t nneedle);
 
 
 #ifndef HAVE_MEMICMP
