@@ -91,6 +91,19 @@ doit ( GpgmeCtx ctx, const char *pattern )
     }
     if ( err != GPGME_EOF )
         fail_if_err (err);
+
+    {
+      char *p;
+
+      p = gpgme_get_op_info (ctx, 0);
+      if (p)
+        {
+          fputs ("<!-- begin operation info -->\n", stdout);
+          fputs (p, stdout);
+          fputs ("<!-- end operation info -->\n", stdout);
+          free (p);
+        }
+    }
 }
 
 
@@ -135,7 +148,7 @@ main (int argc, char **argv )
     GpgmeError err;
     int loop = 0;
     const char *pattern;
-    
+
     if( argc ) {
         argc--; argv++;
     }
@@ -146,7 +159,7 @@ main (int argc, char **argv )
     }
     pattern = argc? *argv : NULL;
 
-    err = gpgme_check_engine();
+    err = gpgme_engine_check_version (GPGME_PROTOCOL_CMS);
     fail_if_err (err);
 
     err = gpgme_new (&ctx);
